/*
 * nvram -- nvram access utility for IBM ppc64 platforms.
 *
 * Copyright (c) 2003 International Business Machines
 * Common Public License Version 0.5 (see COPYRIGHT)
 *
 * ToDo: implement write functions (update firmware vars, etc)
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <netinet/in.h> /* for ntohs */
#define _GNU_SOURCE
#include <getopt.h>

#define MAX_CPUS 64

typedef unsigned short u16;
typedef unsigned int u32;

char *me;
static int verbose;

#define NVRAM_SIG_SP	0x02	/* support processor */
#define NVRAM_SIG_OF	0x50	/* open firmware config */
#define NVRAM_SIG_FW	0x51	/* general firmware */
#define NVRAM_SIG_HW	0x52	/* hardware (VPD) */
#define NVRAM_SIG_SYS	0x70	/* system env vars */
#define NVRAM_SIG_CFG	0x71	/* config data */
#define NVRAM_SIG_ELOG	0x72	/* error log */
#define NVRAM_SIG_VEND	0x7e	/* vendor defined */
#define NVRAM_SIG_FREE	0x7f	/* Free space */
#define NVRAM_SIG_OS	0xa0	/* OS defined */

struct partition_header {
	unsigned char signature;
	unsigned char checksum;
	unsigned short length;
	char name[12];
};
/* Internal representation of NVRAM. */
#define MAX_PARTITIONS 50
struct nvram {
	char *filename;	/* original filename */
	int nparts;	/* number of partitions */
	struct partition_header *parts[MAX_PARTITIONS];	/* points into data */
	int nbytes;	/* size of data in bytes.  This cannot be changed (i.e. hardware size) */
	char data[0];	/* must be at end of this struct */
};

static struct option long_options[] = {
	{"verbose", optional_argument, 0, 'v'},
	{"print-config", optional_argument, 0, 'o'},
	{"print-vpd", optional_argument, 0, 'V'},
	{"print-all-vpd", optional_argument, 0, 'W'},
	{"print-err-log", no_argument, 0, 'e'},
	{"print-event-scan", no_argument, 0, 'E'},
	{"partitions", no_argument, 0, 'p'},
	{"dump", required_argument, 0, 'd'},
	{"nvram-file", required_argument, 0, 'n'},
	{"nvram-size", required_argument, 0, 's'},
	{"help", no_argument, 0, 'h'},
	{0,0,0,0}
};

static void help(void)
{
	printf("nvram options:\n"
	       "  --print-config[=var]	print value of OpenFirmware config variable (or all vars)\n"
	       "  --print-vpd		print VPD\n"
	       "  --print-all-vpd	print VPD including vendor specific data\n"
	       "  --print-err-log	print checkstop error log\n"
	       "  --print-event-scan    print event scan log\n"
	       "  --partitions		print NVRAM paritition header info\n"
	       "  --dump name		raw dump of partition (use --paritions to see names)\n"
	       "  --nvram-file path	specify alternate nvram data file (default /proc/ppc64/nvram\n"
	       "  --nvram-size		specify size of nvram data (for repair operations)\n"
	       "  --verbose (-v)	be (more) verbose\n"
	       "  --help		this help\n"
	       );
}

static int get_of_nvram_size(void)
{
	int fd = open("/proc/device-tree/nvram/#bytes", O_RDONLY);
	int size, len;
	if (fd == -1) {
		perror("cannot open nvram node in device tree");
		return 1024 * 1024;	/* Probably plenty big */
	}
	len = read(fd, &size, sizeof(size));
	close(fd);
	if (len != sizeof(size)) {
		perror("got odd size for nvram node in device tree");
		return 1024 * 1024;	/* Probably plenty big */
	}
	return size;
}

void nvram_read(struct nvram *nvram)
{
	int len, remaining;
	char *p;
	int fd = open(nvram->filename, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "%s: cannot open \"%s\": %s\n", me, nvram->filename, strerror(errno));
		return;
	}
	for (p = nvram->data, remaining = nvram->nbytes;
	     (len = read(fd, p, remaining)) > 0;
	     p += len, remaining -= len)
		/* read it all */;
	if (len == -1) {
		fprintf(stderr, "%s: cannot read \"%s\": %s\n", me, nvram->filename, strerror(errno));
		return;
	}
	close(fd);
	if (remaining) {
		fprintf(stderr, "WARNING:  expected %d bytes, but only read %d!\n", nvram->nbytes, nvram->nbytes - remaining);
		/* preserve the given nbytes...but zero the rest in case someone cares */
		memset(p, 0, remaining);
	}
	if (verbose)
		printf("NVRAM size %d bytes\n", nvram->nbytes);
}

static unsigned char checksum(struct partition_header *p)
{
	unsigned int c_sum, c_sum2;
	unsigned short *sp = (unsigned short *)p->name; /* assume 6 shorts */
	c_sum = p->signature + p->length + sp[0] + sp[1] + sp[2] + sp[3] + sp[4] + sp[5];

	/* The sum probably may have spilled into the 3rd byte.  Fold it back. */
	c_sum = ((c_sum & 0xffff) + (c_sum >> 16)) & 0xffff;
	/* The sum cannot exceed 2 bytes.  Fold it into a checksum */
	c_sum2 = (c_sum >> 8) + (c_sum << 8);
	c_sum = ((c_sum + c_sum2) >> 8) & 0xff;
	return c_sum;
}


/* Map a char for printing in a dump. */
char printmap(char ch)
{
	return isgraph(ch) ? ch : '.';
}

/* Raw data dump of a partition.
 * Note that data_len must be a multiple of 16 bytes which makes
 * for a cheap implementation.
 */
void dump_partition_data(char *data, int data_len)
{
	int blocks = data_len >> 4;	/* 16 byte blocks */
	int i, j;
	char *p = data;
	for (i = 0; i < blocks; i++) {
		printf("   ");
		for (j = 0; j < 16; j++)
			printf("%02x ", p[j]);
		printf("|");
		for (j = 0; j < 16; j++)
			printf("%c", printmap(p[j]));
		printf("|\n");
		p += 16;
	}
}

/* Parse an OpenFirmware common config definition which
 * is of the form name=value and return its length.
 *   data -- points to start of raw data (null terminated).
 *   data_len -- length of remaining data in partition
 *   outname -- 32 char buffer for parsed name
 *   outval -- 4k buffer for output value.
 *
 * Note that the name will always be < 32 chars.  OF does
 * not specify the max value length, but the value is binary
 * and must be unquoted.  We assume 4k is enough.
 *
 */
int parse_of_common(char *data, int data_len, char *outname, char *outval)
{
	char *p, *np;
	char *p_end = data + data_len;

	for (p = data, np = outname; *p && *p != '='; p++) {
		*np++ = *p;
		if (np - outname > 32)
			break;	/* don't overrun */
		if (p >= p_end) {
			printf("partition corrupt:  ran off end parsing name\n");
			return 0;
		}
	}
	*np = '\0';
	if (*p != '=') {
		printf("corrupt data:  no = sign found or name > 31 chars\n");
		return 0;
	}
	p++;
	/* Value needs to be unquoted */
	for (np = outval; *p; p++) {
		if (*p == 0xff) {
			char ch, num;
			p++;
			if (p >= p_end) {
				printf("partition corrupt:  ran off end parsing quoted value\n");
				return 0;
			}
			num = *p & 0x7f;
			ch = (*p & 0x80) ? 0xff : 0;	/* hi bit chooses ff or 00. */
			if (np + num - outval > 4096)
				break;	/* don't overrun */
			/* repeat char */
			while (num > 0) {
				*np++ = ch;
				num--;
			}
		} else {
			*np++ = *p;
			if (np - outval > 4096)
				break;	/* don't overrun */
		}
		if (p >= p_end) {
			printf("partition corrupt:  ran off end parsing value\n");
			return 0;
		}
	}
	*np = '\0';
	if (*p) {
		printf("data value too long for this utility (>4k)\n");
		/* ToDo: recover */
	}
	return p - data;
	
}

/* Dump OpenFirmware common partition.
 * See section 12.4 of the PowerPC CHRP Binding to OpenFirmware
 */
void dump_of_common(char *data, int data_len)
{
	char *p = data;
	char *p_end = data + data_len;
	int len;
	char name[32];
	char val[4096];

	while (*p && p < p_end) {
		/* p is a string name=value */
		len = parse_of_common(p, p_end - p, name, val);
		if (len == 0)
			return;	/* error -- message already printed */
		printf("%s = \"%s\"\n", name, val);
		p += len + 1;
	}
	printf("%d bytes remain in this partition\n", (data + data_len) - p);
}

void dump_partitions(struct nvram *nvram)
{
	char *part_start = nvram->data;
	struct partition_header phead;
	char *name, *p_end;
	unsigned char c_sum;
	char *data_start;
	int data_len;

	printf("Sig Chk  Len  Name\n");
	while (part_start + sizeof(phead) <= nvram->data + nvram->nbytes) {
		memcpy(&phead, part_start, sizeof(phead));
		name = phead.signature == NVRAM_SIG_FREE ? "<free space>" : phead.name;
		printf(" %02x  %02x  %04x %.12s\n", phead.signature, phead.checksum, phead.length, name);
		/* checksum the data */
		p_end = part_start + phead.length * 16;	/* length is in 16 byte blocks */
		c_sum = checksum(&phead);
		if (c_sum != phead.checksum)
			printf("  WARNING: this partition checksum should be %02x!\n", c_sum);
		data_start = part_start + sizeof(phead);
		data_len = p_end - data_start;
		switch (phead.signature) {
		    case NVRAM_SIG_FREE:
			break;	/* no dump */
		    case NVRAM_SIG_SP:
		    case NVRAM_SIG_OF:
		    case NVRAM_SIG_HW:
		    case NVRAM_SIG_FW:
		    case NVRAM_SIG_CFG:
		    case NVRAM_SIG_VEND:
		    case NVRAM_SIG_OS:
		    case NVRAM_SIG_SYS:
		    case NVRAM_SIG_ELOG:
		    default:
			dump_partition_data(data_start, data_len);
			break;
		}
		part_start = p_end;
	}

}

/* Fill in the partition parts of the struct nvram.
 * This makes handling partitions easier for the rest of the code.
 *
 * The spec says that partitions are made up of 16 byte blocks and
 * the partition header must be 16 bytes.  We verify that here.
 */
static void nvram_parse_partitions(struct nvram *nvram)
{
	char *p_start, *p_end;
	struct partition_header *phead;
	unsigned char c_sum;

	if (sizeof(struct partition_header) != 16) {
		fprintf(stderr, "%s: INTERNAL ERROR: partition_header struct is not 16 bytes\n", me);
		exit(1);
	}
	for (p_start = nvram->data;
	     p_start + 16 <= nvram->data + nvram->nbytes;
	     p_start = p_end) {
		phead = (struct partition_header *)p_start;
		nvram->parts[nvram->nparts++] = phead;
		c_sum = checksum(phead);
		if (c_sum != phead->checksum)
			fprintf(stderr, "  ERRROR: this partition checksum should be %02x!\n", c_sum);
		p_end = p_start + phead->length * 16;	/* length is in 16 byte blocks */
	}
	if (verbose)
		printf("NVRAM contains %d partitions\n", nvram->nparts);
}

/* Find a parition given a signature and name.
 * If signature is zero (invalid) it is not used for matching.
 * If name is NULL it is ignored.
 * start is the partition to to resume a search (NULL starts at the beginning)
 */
static struct partition_header *nvram_find_partition(unsigned char signature, char *name, struct partition_header *start, struct nvram *nvram)
{
	struct partition_header *phead;
	int i;

	/* Get starting partition.  This is not terribly efficient... */
	if (start == NULL) {
		i = 0;
		if (verbose > 1)
			printf("find partition starts with zero\n");
	} else {
		for (i = 0; i < nvram->nparts; i++)
			if (nvram->parts[i] == start)
				break;
		i++;	/* start at next partition */
		if (verbose > 1)
			printf("find partition starts with %d\n", i);
	}

	/* Search starting with partition i... */
	while (i < nvram->nparts) {
		phead = nvram->parts[i];
		if (signature == '\0' || signature == phead->signature) {
			if (name == NULL || strncmp(name, phead->name, sizeof(phead->name)) == 0) {
				return phead;
			}
		}
		i++;
	}
	return NULL;
}

static void print_partition_table(struct nvram *nvram)
{
	struct partition_header *phead;
	int i = 0;

	printf(" # Sig Chk  Len  Name\n");
	for (phead = nvram_find_partition(0, NULL, NULL, nvram);
	     phead != NULL;
	     phead = nvram_find_partition(0, NULL, phead, nvram)) {
		printf("%2d  %02x  %02x  %04x %.12s\n", i++, phead->signature, phead->checksum, phead->length, phead->name);
	}
}

/* Copy a value into a buf.
 * The first two bytes of the value is a length.
 * Return pointer to byte after the value.
 */
static char *getvalue(char *p, char *buf)
{
	int len = *p++;
	len |= ((*p++) << 8);
	memcpy(buf, p, len);
	buf[len] = '\0';
	return p+len;
}

/* Copy a value into a buf.
 * The first one bytes of the value is a length.
 * Return pointer to byte after the value.
 */
static char *getsmallvalue(char *p, char *buf)
{
	int len = *p++;
	memcpy(buf, p, len);
	buf[len] = '\0';
	return p+len;
}

/* Lookup a VPD field name (always 2 chars) and return a human readable string.
 * Return NULL if it is not a standard field.
 */
static char *lookupfield(char *p)
{
	static struct { char *name; char *desc; } descs[] = {
		{"PN", "Part Number"},
		{"FN", "FRU Number"},
		{"EC", "EC Level"},
		{"MN", "Manufacture ID"},
		{"SN", "Serial Number"},
		{"LI", "Load ID"},
		{"RL", "ROM Level"},
		{"RM", "Alterable ROM Level"},
		{"NA", "Network Address"},
		{"DD", "Device Driver Level"},
		{"DG", "Diagnostic Level"},
		{"LL", "Loadable Microcode Level"},
		{"VI", "Vendor ID/Device ID"},
		{"FU", "Function Number"},
		{"SI", "Subsystem Vendor ID/Device ID"},
		{"VK", "Platform"},		/* "RS6K" => VPD is present */
		{"TM", "Model"},		/* IBM specific? */
		{"YL", "Location Code"},	/* IBM specific? */
		{"BR", "Brand"},		/* IBM specific */
		{"CI", "CEC ID"},		/* IBM specific */
		{"RD", "Rack ID"},		/* IBM specific */
		{"PA", "Op Panel Installed"},	/* IBM specific */
		{"NN", "Node Name"},		/* IBM specific */
	};
	int i;

	for (i = 0; (i < sizeof(descs) / sizeof(descs[0])); i++) {
		if (strcmp(p, descs[i].name) == 0)
			return descs[i].desc;
	}
	return NULL;
}

/* Format and print a VPD field and return a ptr to the next field. */
static char *printVPDfield(char *p, int show_all)
{
	char field[3];
	char value[256];
	char *fname;

	field[0] = p[0]; field[1] = p[1]; field[2] = 0;
	p+=2;
	p = getsmallvalue(p, value);
	if ((fname = lookupfield(field)) != NULL)
		printf("	%-20s %s\n", fname, value);
	else if (show_all)
		printf("	%-20s %s\n", field, value);
	return p;
}

/* ToDo: use vpd.c stuff. */
/* Dump Vital Product Data
 * See Chapter 18: Expansion ROMs of the PCI spec.
 */
void dump_vpd(struct nvram *nvram, int show_all)
{
	struct partition_header *phead;
	char *p, *p_end;
	char value[4096];

	phead = nvram_find_partition(NVRAM_SIG_HW, "ibm,vpd", NULL, nvram);
	if (!phead) {
		fprintf(stderr, "%s: there is no ibm,vpd partition!\n", me);
		return;
	}
	p = (char *)(phead + 1);
	p_end = (char *)(phead + phead->length);
	while (*p && p < p_end) {
		if (*p == 0x82) {	/* Identification string descriptor. */
			p++;
			p = getvalue(p, value);
			printf("%s\n", value);	/* print id string */
			while (*p != 0x79) {	/* loop until VPD end tag */
				int vpdlen;
				char *vpd_endp;
				p++;
				vpdlen = *p++;
				vpdlen |= ((*p++) << 8);
				vpd_endp = p + vpdlen;
				while (p < vpd_endp)
					p = printVPDfield(p, show_all);
			}
			p++;
			/* printf("checksum byte=0x%x\n", *p); */
			p++;
		} else if (*p == 0) {	/* end tag */
			break;
		}
	}
	if (*p && p < p_end) {
		printf("found unknown descriptor byte 0x%x\n", *p);
	}
}

#define NUM_ERRLOG_PER_LINE 16
void dump_errlog_data(char *msg, u16 *data, int num)
{
	int i;
	int block, addr;
	int num_blocks = num / NUM_ERRLOG_PER_LINE;
	int num_remain = num % NUM_ERRLOG_PER_LINE;

	printf("%s\n", msg);
	if (num_remain)
		num_blocks++;
	addr = 0;
	for (block = 0; block < num_blocks; block++) {
		printf("%03x:  ", addr);
		for (i = 0; i < NUM_ERRLOG_PER_LINE; i++) {
			printf("%04x ", *data++);
			addr++;
			if (addr >= num)
				break;
		}
		printf("\n");
	}
}


/* Dump ibm,err-log partition which contains checkstop info.
 * ToDo: this code needs more work.
 * See IBM RPA (IBM internal use only -- sorry).
 */
void dump_errlog(struct nvram *nvram)
{
	struct partition_header *phead;
	u16 *p, *p_end;	/* Note: data is organized into 16bit big endian (network byte order) */
	int p_max;	/* max index to go out of bounds of the partition */
	int i, cpu;
	char checkstop_count;
	int offset;
	int num_cpus;
	int num_memctrls;
	int num_ioctrls;
	u16 *sys_regs;	/* System specific registers (e.g. bus arbitration chips, etc */
	u16 *cpu_regs[MAX_CPUS+1];
	u16 *memctrl_data;
	u16 *ioctrl_data;

	phead = nvram_find_partition(NVRAM_SIG_SP, "ibm,err-log", NULL, nvram);
	if (!phead) {
		fprintf(stderr, "%s: there is no ibm,err-log partition!\n", me);
		return;
	}
	p = (u16 *)(phead + 1);
	p_end = (u16 *)(phead + phead->length);
	p_max = p_end - p;	/* max in 16bit values */
	if (p_max < 4) {
		printf("Corrupt ibm,err-log partition in nvram\n");
		return;
	}
	/* index 0 is checkstop count (high byte), semaphores (low byte) */
	i = 0;	/* index through short words */
	checkstop_count = p[i] >> 8;
	if (checkstop_count)
		printf("Checkstops detected: %d\n", checkstop_count);
	else
		printf("No checkstops have been detected.\n");
	/* index 1 is system specific register offset */
	i++;
	offset = ntohs(p[i])/2+1;
	sys_regs = offset + i < p_max ? p + offset + i : 0;

	/* index 2 is number of cpus */
	i++;
	num_cpus = ntohs(p[i]);
	printf("CPUS: %d\n", num_cpus);

	/* Next indexes are offsets to cpu specific regs */
	for (cpu = 0; cpu < num_cpus; cpu++) {
		i++;
		if (cpu < MAX_CPUS) {
			offset = ntohs(p[i])/2+1;
			cpu_regs[cpu] = offset + i < p_max ? p + offset + i : 0;
		}
	}
	if (num_cpus > MAX_CPUS)
		num_cpus = MAX_CPUS;	/* just in case... */

	/* next index is number of memory controllers */
	i++;
	num_memctrls = ntohs(p[i]);
	printf("Memory Controllers: %d\n", num_memctrls);

	/* next index is offset of memory controller data */
	i++;	/* ToDo: this may be a list of offsets...manual doesn't show that but only 1 seems odd */
	offset = ntohs(p[i])/2+1;
	memctrl_data = offset + i < p_max ? p + offset + i : 0;

	/* next index is number of I/O Subsystem controllers */
	i++;
	num_ioctrls = ntohs(p[i]);
	printf("I/O Controllers: %d\n", num_ioctrls);

	/* next index is offset of I/O Subsystem controller data */
	i++;	/* ToDo: this may be a list of offsets...manual doesn't show that but only 1 seems odd */
	offset = ntohs(p[i])/2+1;
	ioctrl_data = offset + i < p_max ? p + offset + i : 0;

	/* End of header */

	/* Now dump sections collected by the header. */
	if (sys_regs && num_cpus > 0) {
		/* ToDo: what is the length of the data?  We dump until the first cpu data. */
		dump_errlog_data("System Specific Registers", sys_regs, cpu_regs[0] - sys_regs);
	}
	cpu_regs[num_cpus] = ioctrl_data; /* artificial "next cpu" data for length */
	for (cpu = 0; cpu < num_cpus; cpu++) {
		char buf[64];
		int len;
		/* ToDo: what is the length of the data?  We dump until the next cpu data. */
		len = cpu_regs[cpu+1] - cpu_regs[cpu];
		sprintf(buf, "CPU %d Register Data (len=%x, offset=%x)", cpu, len, cpu_regs[cpu]-p);
		if (len < 4096)	/* reasonable bound */
			dump_errlog_data(buf, cpu_regs[cpu], len);
	}
}

/* Format a BCD date/time to human readable form.  Returns ptr to static buffer overwritten on next call */
static char *fmt_BCDdate(u32 BCDdate, u32 BCDtime)
{
	static char buf[128];
	struct tm tm, *gtm;
	time_t thetime;

	memset((void *)&tm, 0, sizeof(tm));

	/* BCDtime HHMMSS00 */
	tm.tm_sec  = ((BCDtime & 0x0000f000) >> 12) * 10 + ((BCDtime & 0x00000f00) >> 8);
	tm.tm_min  = ((BCDtime & 0x00f00000) >> 20) * 10 + ((BCDtime & 0x000f0000) >> 16);
	tm.tm_hour = ((BCDtime & 0xf0000000) >> 28) * 10 + ((BCDtime & 0x0f000000) >> 24);

	/* BCDdate YYYYMMDD */
	tm.tm_year = (((BCDdate & 0xf0000000) >> 28) * 1000 + ((BCDdate & 0x0f000000) >> 24) * 100 +
	              ((BCDdate & 0x00f00000) >> 20) * 10   + ((BCDdate & 0x000f0000) >> 16)) - 1900;
	tm.tm_mon  = ((BCDdate & 0x0000f000) >> 12) * 10 + ((BCDdate & 0x00000f00) >> 8) - 1;
	tm.tm_mday = ((BCDdate & 0x000000f0) >> 4) * 10  +  (BCDdate & 0x0000000f);
	thetime = mktime(&tm);
	gtm = localtime(&thetime);	/* This fills in the weekday, etc. */

	/* Detect corrupt date */
	if (tm.tm_sec > 59 || tm.tm_min > 59 || tm.tm_hour > 23 || tm.tm_year < 95 || tm.tm_year > 200 || tm.tm_mon > 12 || tm.tm_mday > 31)
		sprintf(buf, "invalid time:%08x date:%08x", BCDtime, BCDdate);
	else
		strftime(buf, sizeof(buf), "%Y GMT", gtm);
	return buf;
}

/* Dump event-scan data.
 */
void dump_log_entry(int logtype, int flags, char *data, int len)
{
	struct rtas_error {
		int version:8;
		int severity:3;
		int rtas_disposition:2;
		int optional_part_presence:1;
		int reserved1:2;
		int initiator:4;
		int target:4;
		int errtype:8;
		int extended_error_length:32;
	} *err;
	char *severity_names[] = {
		"NO ERROR", "EVENT", "WARNING", "ERROR_SYNC", "ERROR", "FATAL", "(6)", "(7)"
	};
	char *rtas_disposition_names[] = {
		"FULLY RECOVERED", "LIMITED RECOVERY", "NOT RECOVERED", "(4)"
	};
	char *entity_names[] = { /* for initiator & targets */
		"UNKNOWN", "CPU", "PCI", "ISA", "MEMORY", "POWER MANAGEMENT", "HOT PLUG", "(7)", "(8)",
		"(9)", "(10)", "(11)", "(12)", "(13)", "(14)", "(15)"
	};

	err = (struct rtas_error *)data;	/* ToDo: endian bug here */
	printf("\tSeverity: %s\n", severity_names[err->severity]);
	printf("\tDisposition: %s\n", rtas_disposition_names[err->rtas_disposition]);
	if (err->initiator) printf("\tInitiator: %s\n", entity_names[err->initiator]);
	if (err->target) printf("\tTarget: %s\n", entity_names[err->target]);
	printf("\tExtended error type: %d\n", err->errtype);
	if (err->optional_part_presence && err->version != 3) {
		/* hack to dump non-version 3 info */
		dump_partition_data((char *)(err+1), 28);
	}

	if (err->optional_part_presence && err->version == 3) {	/* only support extended info for version 3 */
		struct exthdr {
			unsigned int valid:1;
			unsigned int unrecoverable:1;
			unsigned int recoverable:1;
			unsigned int unrecoverable_bypassed:1;	/* i.e. degraded performance */
			unsigned int predictive:1;
			unsigned int newlog:1;
			unsigned int bigendian:1;		/* always 1 */
			unsigned int reserved1:1;

			unsigned int platform_specific:1;
			unsigned int reserved2:3;
			unsigned int platform_value:4;		/* valid iff platform_specific */

			unsigned int power_pc:1;			/* always 1 */
			unsigned int reserved3:2;
			unsigned int addr_invalid:1;		/* failing_address is invalid */
			unsigned int format_type:4;
#define EXTLOG_FMT_CPU 1
#define EXTLOG_FMT_MEMORY 2
#define EXTLOG_FMT_IO 3
#define EXTLOG_FMT_POST 4
#define EXTLOG_FMT_ENV 5
#define EXTLOG_FMT_POW 6
#define EXTLOG_FMT_IBMDIAG 12
#define EXTLOG_FMT_IBMSP 13

			unsigned int non_hardware:1;		/* Firmware or software is suspect */
			unsigned int hot_plug:1;			/* Failing component may be hot plugged */
			unsigned int group_failure:1;		/* Group of components should be replaced */
			unsigned int reserved4:1;
			unsigned int residual:1;			/* Residual error from previous boot (maybe a crash) */
			unsigned int boot:1;			/* Error during boot */
			unsigned int config_change:1;		/* Configuration changed since last boot */
			unsigned int post:1;			/* Error during POST */

			unsigned int bcdtime:32;			/* Time of error in BCD HHMMSS00 */
			unsigned int bcddate:32;			/* Time of error in BCD YYYYMMDD */
		} *exthdr = (struct exthdr *)(err+1);	/* ToDo: endian error here */

		if (exthdr->valid) {
			printf("\tStatus:%s%s%s%s%s\n",
			       exthdr->unrecoverable ? " unrecoverable" : "",
			       exthdr->recoverable ? " recoverable" : "",
			       exthdr->unrecoverable_bypassed ? " bypassed" : "",
			       exthdr->predictive ? " predictive" : "",
			       exthdr->newlog ? " new" : "");
			printf("\tDate:  %s\n", fmt_BCDdate(exthdr->bcddate, exthdr->bcdtime));
			switch (exthdr->format_type) {
			    case EXTLOG_FMT_CPU: {
					struct cpuhdr {
						unsigned int internal:1;
						unsigned int intcache:1;
						unsigned int extcache_parity:1;	/* or multi-bit ECC */
						unsigned int extcache_ecc:1;
						unsigned int sysbus_timeout:1;
						unsigned int io_timeout:1;
						unsigned int sysbus_parity:1;
						unsigned int sysbus_protocol:1;
						unsigned int cpuid:8;
						unsigned int element:16;
						unsigned int failing_address_hi:32;
						unsigned int failing_address_lo:32;
						/* 16 bytes reserved here */
					} *cpuhdr = (struct cpuhdr *)(exthdr+1);
					printf("\tCPU detected failure\n");
					if (cpuhdr->internal) printf("\tInternal error (not cache)\n");
					if (cpuhdr->intcache) printf("\tInternal cache\n");
					if (cpuhdr->extcache_parity) printf("\tExternal cache parity (or multi-bit)\n");
					if (cpuhdr->extcache_ecc) printf("\tExternal cache ECC\n");
					if (cpuhdr->sysbus_timeout) printf("\tSystem bus timeout\n");
					if (cpuhdr->io_timeout) printf("\tI/O timeout\n");
					if (cpuhdr->sysbus_parity) printf("\tSystem bus parity\n");
					if (cpuhdr->sysbus_protocol) printf("\tSystem bus protocol/transfer\n");
					printf("\tCPU id: %d\n", cpuhdr->cpuid);
					printf("\tFailing element: 0x%04x\n", cpuhdr->element);
					if (!exthdr->addr_invalid)
						printf("\tFailing address: 0x%08x%08x\n", cpuhdr->failing_address_hi, cpuhdr->failing_address_lo);
					break;
				}
			    case EXTLOG_FMT_MEMORY: {
					struct memhdr {
						unsigned int uncorrectable:1;
						unsigned int ECC:1;
						unsigned int threshold_exceeded:1;
						unsigned int control_internal:1;
						unsigned int bad_address:1;
						unsigned int bad_data:1;
						unsigned int bus:1;
						unsigned int timeout:1;
						unsigned int sysbus_parity:1;
						unsigned int sysbus_timeout:1;
						unsigned int sysbus_protocol:1;
						unsigned int hostbridge_timeout:1;
						unsigned int hostbridge_parity:1;
						unsigned int reserved1:1;
						unsigned int support:1;
						unsigned int sysbus_internal:1;
						unsigned int mem_controller_detected:8;	/* who detected fault? */
						unsigned int mem_controller_faulted:8;	/* who caused fault? */
						unsigned int failing_address_hi:32;
						unsigned int failing_address_lo:32;
						unsigned int ecc_syndrome:16;
						unsigned int memory_card:8;
						unsigned int reserved2:8;
						unsigned int sub_elements:32;		/* one bit per element */
						unsigned int element:16;
					} *memhdr = (struct memhdr *)(exthdr + 1);
					printf("\tMEMORY detected failure\n");
					if (memhdr->uncorrectable) printf("\tUncorrectable Memory error\n");
					if (memhdr->ECC) printf("\tECC Correctable error\n");
					if (memhdr->threshold_exceeded) printf("\tCorrectable threshold exceeded\n");
					if (memhdr->control_internal) printf("\tMemory Controller internal error\n");
					if (memhdr->bad_address) printf("\tMemory Address error\n");
					if (memhdr->bad_data) printf("\tMemory Data error\n");
					if (memhdr->bus) printf("\tMemory bus/switch internal error\n");
					if (memhdr->timeout) printf("\tMemory timeout\n");
					if (memhdr->sysbus_parity) printf("\tSystem bus parity\n");
					if (memhdr->sysbus_timeout) printf("\tSystem bus timeout\n");
					if (memhdr->sysbus_protocol) printf("\tSystem bus protocol/transfer\n");
					if (memhdr->hostbridge_timeout) printf("\tI/O Host Bridge timeout\n");
					if (memhdr->hostbridge_parity) printf("\tI/O Host Bridge parity\n");
					if (memhdr->support) printf("\tSystem support function error\n");
					if (memhdr->sysbus_internal) printf("\tSystem bus internal hardware/switch error\n");
					printf("\tMemory Controller that detected failure: %d\n", memhdr->mem_controller_detected);
					printf("\tMemory Controller that faulted: %d\n", memhdr->mem_controller_faulted);
					if (!exthdr->addr_invalid)
						printf("\tFailing address: 0x%016x%016x\n", memhdr->failing_address_hi, memhdr->failing_address_lo);
					printf("\tECC syndrome bits: 0x%04x\n", memhdr->ecc_syndrome);
					printf("\tMemory Card: %d\n", memhdr->memory_card);
					printf("\tFailing element: 0x%04x\n", memhdr->element);
					printf("\tSub element bits: 0x%08x\n", memhdr->sub_elements);
					break;
				}
			    case EXTLOG_FMT_IO: {
					struct iohdr {
						unsigned int bus_addr_parity:1;
						unsigned int bus_data_parity:1;
						unsigned int bus_timeout:1;
						unsigned int bridge_internal:1;
						unsigned int non_pci:1;		/* i.e. secondary bus such as ISA */
						unsigned int mezzanine_addr_parity:1;
						unsigned int mezzanine_data_parity:1;
						unsigned int mezzanine_timeout:1;

						unsigned int bridge_via_sysbus:1;
						unsigned int bridge_via_mezzanine:1;
						unsigned int bridge_via_expbus:1;
						unsigned int detected_by_expbus:1;
						unsigned int expbus_data_parity:1;
						unsigned int expbus_timeout:1;
						unsigned int expbus_connection_failure:1;
						unsigned int expbus_not_operating:1;

						/* IOA signalling the error */
						unsigned int pci_sig_busno:8;
						unsigned int pci_sig_devfn:8;
						unsigned int pci_sig_deviceid:16;
						unsigned int pci_sig_vendorid:16;
						unsigned int pci_sig_revisionid:8;
						unsigned int pci_sig_slot:8;	/* 00 => system board, ff => multiple */

						/* IOA sending at time of error */
						unsigned int pci_send_busno:8;
						unsigned int pci_send_devfn:8;
						unsigned int pci_send_deviceid:16;
						unsigned int pci_send_vendorid:16;
						unsigned int pci_send_revisionid:8;
						unsigned int pci_send_slot:8;	/* 00 => system board, ff => multiple */

					} *iohdr = (struct iohdr *)(exthdr+1);

					printf("\tI/O detected failure\n");
					if (iohdr->bus_addr_parity) printf("\tI/O bus address parity\n");
					if (iohdr->bus_data_parity) printf("\tI/O bus data parity\n");
					if (iohdr->bus_timeout) printf("\tI/O bus timeout, access or other\n");
					if (iohdr->bridge_internal) printf("\tI/O bus bridge/device internal\n");
					if (iohdr->non_pci) printf("\tSignaling IOA is a PCI to non-PCI bridge (e.g. ISA)\n");
					if (iohdr->mezzanine_addr_parity) printf("\tMezzanine/System bus address parity\n");
					if (iohdr->mezzanine_data_parity) printf("\tMezzanine/System bus data parity\n");
					if (iohdr->mezzanine_timeout) printf("\tMezzanine/System bus timeout, transfer or protocol\n");
					if (iohdr->bridge_via_sysbus) printf("\tBridge is connected to system bus\n");
					if (iohdr->bridge_via_mezzanine) printf("\tBridge is connected to memory controller via mezzanine bus\n");
					if (iohdr->bridge_via_expbus) printf("\tBridge is connected to I/O expansion bus\n");
					if (iohdr->detected_by_expbus) printf("\tError on system bus detected by I/O expansion bus controller\n");
					if (iohdr->expbus_data_parity) printf("\tI/O expansion bus data error\n");
					if (iohdr->expbus_timeout) printf("\tI/O expansion bus timeout, access or other\n");
					if (iohdr->expbus_connection_failure) printf("\tI/O expansion bus connection failure\n");
					if (iohdr->expbus_not_operating) printf("\tI/O expansion unit not in an operating state (powered off, off-line)\n");

					printf("\tIOA Signaling the error: %d:%d.%d vendor:%04x device:%04x rev:%02x slot:%d\n",
					       iohdr->pci_sig_busno, iohdr->pci_sig_devfn >> 3, iohdr->pci_sig_devfn & 0x7,
					       iohdr->pci_sig_vendorid, iohdr->pci_sig_deviceid, iohdr->pci_sig_revisionid, iohdr->pci_sig_slot);
					printf("\tIOA Sending during the error: %d:%d.%d vendor:%04x device:%04x rev:%02x slot:%d\n",
					       iohdr->pci_send_busno, iohdr->pci_send_devfn >> 3, iohdr->pci_send_devfn & 0x7,
					       iohdr->pci_send_vendorid, iohdr->pci_send_deviceid, iohdr->pci_send_revisionid, iohdr->pci_send_slot);

					break;
				}
			    case EXTLOG_FMT_POST:
				printf("\tPOST detected failure\n");
				break;
			    case EXTLOG_FMT_ENV:
				printf("\tEnvironment and Power Warning\n");
				break;
			    case EXTLOG_FMT_POW:
				printf("\tPower Management Event\n");
				break;
			    case EXTLOG_FMT_IBMDIAG:
				printf("\tIBM Diagnostic Log\n");
				break;
			    case EXTLOG_FMT_IBMSP: {
					struct sphdr {
						unsigned int ibm:32;	/* "IBM\0" */

						unsigned int timeout:1;
						unsigned int i2c_bus:1;
						unsigned int i2c_secondary_bus:1;
						unsigned int sp_memory:1;
						unsigned int sp_registers:1;
						unsigned int sp_communication:1;
						unsigned int sp_firmware:1;
						unsigned int sp_hardware:1;

						unsigned int vpd_eeprom:1;
						unsigned int op_panel:1;
						unsigned int power_controller:1;
						unsigned int fan_sensor:1;
						unsigned int thermal_sensor:1;
						unsigned int voltage_sensor:1;
						unsigned int reserved1:2;

						unsigned int serial_port:1;
						unsigned int nvram:1;
						unsigned int rtc:1;
						unsigned int jtag:1;
						unsigned int tod_battery:1;
						unsigned int reserved2:1;
						unsigned int heartbeat:1;
						unsigned int surveillance:1;

						unsigned int pcn_connection:1;	/* power control network */
						unsigned int pcn_node:1;
						unsigned int reserved3:2;
						unsigned int pcn_access:1;
						unsigned int reserved:3;

						unsigned int sensor_token:32;	/* zero if undef */
						unsigned int sensor_index:32;	/* zero if undef */
					} *sphdr = (struct sphdr *)(exthdr+1);
					u32 eyecatcher;
					printf("\tIBM Service Processor Log\n");
					eyecatcher = sphdr->ibm;
					if (strcmp((char *)&eyecatcher, "IBM") != 0)
						printf("\tThis log entry may be corrupt (IBM signature malformed)\n");
					if (sphdr->timeout) printf("\tTimeout on communication response from service processor\n");
					if (sphdr->i2c_bus) printf("\tI2C general bus error\n");
					if (sphdr->i2c_secondary_bus) printf("\tI2C secondary bus error\n");
					if (sphdr->sp_memory) printf("\tInternal service processor memory error\n");
					if (sphdr->sp_registers) printf("\tService processor error accessing special registers\n");
					if (sphdr->sp_communication) printf("\tService processor reports unknown communcation error\n");
					if (sphdr->sp_firmware) printf("\tInternal service processor firmware error\n");
					if (sphdr->sp_hardware) printf("\tOther internal service processor hardware error\n");
					if (sphdr->vpd_eeprom) printf("\tService processor error accessing VPD EEPROM\n");
					if (sphdr->op_panel) printf("\tService processor error accessing Operator Panel\n");
					if (sphdr->power_controller) printf("\tService processor error accessing Power Controller\n");
					if (sphdr->fan_sensor) printf("\tService processor error accessing Fan Sensor\n");
					if (sphdr->thermal_sensor) printf("\tService processor error accessing Thermal Sensor\n");
					if (sphdr->voltage_sensor) printf("\tService processor error accessing Voltage Sensor\n");
					if (sphdr->serial_port) printf("\tService processor error accessing serial port\n");
					if (sphdr->nvram) printf("\tService processor detected NVRAM error\n");
					if (sphdr->rtc) printf("\tService processor error accessing real time clock\n");
					if (sphdr->jtag) printf("\tService processor error accessing JTAG/COP\n");
					if (sphdr->tod_battery) printf("\tService processor or RTAS detects loss of voltage from TOD battery\n");
					if (sphdr->heartbeat) printf("\tLoss of heartbeat from Service processor\n");
					if (sphdr->surveillance) printf("\tService processor detected a surveillance timeout\n");
					if (sphdr->pcn_connection) printf("\tPower Control Network general connection failure\n");
					if (sphdr->pcn_node) printf("\tPower Control Network node failure\n");
					if (sphdr->pcn_access) printf("\tService processor error accessing Power Control Network\n");

					if (sphdr->sensor_token) printf("\tSensor Token 0x%08x (%d)\n", sphdr->sensor_token, sphdr->sensor_token);
					if (sphdr->sensor_index) printf("\tSensor Index 0x%08x (%d)\n", sphdr->sensor_index, sphdr->sensor_index);
					break;
				}
			    default:
				printf("\tUnknown extended error format %d\n", exthdr->format_type);
				break;
			}
		}
		
	}
	
}


/* Dump ibm,es-logs partition contains a service processor log
 * See IBM RPA (IBM internal use only -- sorry).
 */
void dump_eventscanlog(struct nvram *nvram)
{
	struct partition_header *phead;
	u32 *p, *p_end;	/* Note: data is organized into 32bit big endian (network byte order) */
	int p_max;	/* max index to go out of bounds of the partition */
	int lognum;
	int num_logs;
#define MAX_EVENTLOGS 100
	u32 loghdr[MAX_EVENTLOGS+1];

	phead = nvram_find_partition(NVRAM_SIG_SP, "ibm,es-logs", NULL, nvram);
	if (!phead) {
		fprintf(stderr, "%s: there is no ibm,es-logs partition!\n", me);
		return;
	}
	p = (u32 *)(phead + 1);
	p_end = (u32 *)(phead + phead->length);
	p_max = p_end - p;	/* max in 32bit values */
	if (p_max < 1) {
		printf("Corrupt ibm,es-logs partition in nvram\n");
		return;
	}
	num_logs = ntohl(*p);
	printf("Number of Logs: %d\n", num_logs);
	if (num_logs > MAX_EVENTLOGS) {
		num_logs = MAX_EVENTLOGS;
		fprintf(stderr, "%s: limiting to %d log entries (program limit)\n", me, num_logs);
	}
	if (num_logs > p_max-1) {
		/* of course this leaves no room for log data (i.e. corrupt partition) */
		num_logs = p_max-1;
		fprintf(stderr, "%s: limiting to %d log entries (partition limit)\n", me, num_logs);
	}
	
	for (lognum = 0; lognum < num_logs; lognum++) {
		loghdr[lognum] = ntohl(p[lognum+1]);
	}
	/* artificial log entry (offset) to put a limit on the last log */
	loghdr[num_logs] = p_max * sizeof(u32);

	for (lognum = 0; lognum < num_logs; lognum++) {
		u32 hdr = loghdr[lognum];
		int flags = (hdr >> 24) & 0xff;
		int logtype = (hdr >> 16) & 0xff;
		int start = hdr & 0xffff;
		int end = loghdr[lognum+1] & 0xffff;
		printf("Log Entry %d:  flags: 0x%02x  type: 0x%02x\n", lognum, flags, logtype);
		dump_log_entry(logtype, flags, ((char *)p)+start, end-start);
	}
}


/* Dump raw data of a partition.  Mainly for debugging.
 */
void dump_raw_partition(struct nvram *nvram, char *name)
{
	struct partition_header *phead;
	char *p, *p_end;

	phead = nvram_find_partition(0, name, NULL, nvram);
	if (!phead) {
		fprintf(stderr, "%s: there is no %s partition!\n", me, name);
		return;
	}
	p = (char *)(phead + 1);
	p_end = (char *)(phead + phead->length);
	dump_partition_data(p, p_end - p);
}

/* Print a single OF var...or all if "" is used */
static void print_of_config_var(char *of_config_var, struct nvram *nvram)
{
	struct partition_header *phead;
	int varlen;
	char *p, *p_next, *data_start, *data_end;

	phead = nvram_find_partition(NVRAM_SIG_SYS, "common", NULL, nvram);
	if (!phead) {
		fprintf(stderr, "%s: there is no open firmware common partition!\n", me);
		return;
	}
	data_start = (char *)(phead + 1);
	data_end = (char *)(phead + phead->length);

	varlen = strlen(of_config_var);
	for (p = data_start;
	     (*p != '\0') && (p + varlen < data_end);
	     p = p_next) {
		/* ToDo: unquote the 0xff N sequence and re-quote the value in \xFF notation
		 * in case binary data is found.  For now don't bother.
		 */
		if (varlen == 0) {
			printf("%s\n", p);
		} else if ((p[varlen] == '=') && strncmp(of_config_var, p, varlen) == 0) {
			printf("%s\n", p+varlen+1);
		}
		p_next = p + strlen(p) + 1;
	}
}

int main (int argc, char *argv[])
{
	int ret;
	struct nvram *nvram;
	int nvram_size = 0;
	char *nvram_filename = "/dev/nvram";
	int option_index;
	char *endp;
	int of_nvram_size;
	char *of_config_var = NULL;
	int print_partitions = 0;
	int print_vpd = 0;
	int print_errlog = 0;
	int print_event_scan = 0;
	char *dump_name = NULL;

	me = argv[0];
	if (argc == 1) {
		help();
		exit(1);
	}
	for (;;) {
		option_index = 0;
		ret = getopt_long_only(argc, argv, "v::", long_options, &option_index);
		if (ret == -1)
			break;
		switch (ret) {
		    case 'h':
			help();
			exit(0);
		    case 'v':
			verbose += (optarg ? atoi(optarg) : 1);
			break;
		    case 'd':
			dump_name = optarg;
			break;
		    case 'n':
			nvram_filename = optarg;
			break;
		    case 'o':
			of_config_var = optarg ? optarg : "";	/* empty string is flag for "all" */
			break;
		    case 'p':
			print_partitions = 1;
			break;
		    case 's':
			nvram_size = strtoul(optarg, &endp, 10);
			if (!*optarg || *endp) {
				fprintf(stderr, "specify nvram-size as an integer\n");
				exit(1);
			}
			break;
		    case 'V':
			print_vpd = 1;
			break;
		    case 'W':
			print_vpd = 2;
			break;
		    case 'e':
			print_errlog = 1;
			break;
		    case 'E':
			print_event_scan = 1;
			break;
		    case '?':
			exit(1);
			break;
		    default:
			printf("huh?\n");
			break;
		}
	}
	of_nvram_size = get_of_nvram_size();
	if (nvram_size == 0) {
		struct stat sb;

		if (stat(nvram_filename, &sb) == -1) {
			fprintf(stderr, "%s: cannot stat %s: %s\n", me, nvram_filename, strerror(errno));
			exit(2);
		}
		nvram_size = sb.st_size ? sb.st_size : of_nvram_size;
	}
	if (nvram_size != get_of_nvram_size()) {
		fprintf(stderr, "WARNING:  specified nvram size %d does not match this machine %d!\n", nvram_size, of_nvram_size);
	}
	nvram = (struct nvram *)malloc(sizeof(struct nvram) + nvram_size);
	if (!nvram) {
		fprintf(stderr, "%s: cannot allocate space for nvram of %d bytes\n", me, nvram_size);
		return 1;
	}
	nvram->nbytes = nvram_size;
	nvram->filename = nvram_filename;
	nvram_read(nvram);
	if (nvram->nbytes == 0) {
		fprintf(stderr, "no nvram data\n");
		return 1;	/* no data */
	}
	nvram_parse_partitions(nvram);

	if (print_partitions)
		print_partition_table(nvram);
	if (of_config_var)
		print_of_config_var(of_config_var, nvram);
	if (print_vpd)
		dump_vpd(nvram, print_vpd == 2);
	if (print_errlog)
		dump_errlog(nvram);
	if (print_event_scan)
		dump_eventscanlog(nvram);
	if (dump_name)
		dump_raw_partition(nvram, dump_name);
	// dump_partitions(nvram);
	return 0;
}
